function [gamma, Ahat, logl, dens] = hmm_post (X, A, pi0, mu, Sigma, QUIET)

%hmm_post  Implements backward Markovian smoothing (forward-backward alternative).
%	Use: [gamma,Ahat,logl,dens] = hmm_post(X,A,pi0,mu,Sigma) where
%       gamma contains the smoothing probabilities and Ahat the cumulated
%       bivariate smoothing probabilities (normalized so as to give the EM
%       re-estimate for A). logl is the log-likelihood and dens contains
%       the values of all Gaussian densities for each time index.

% H2M Toolbox, $Revision: 2.3 $
% Olivier Capp, 06/12/2002 - $Date: 2006/06/13 11:51:19 $
% http://www.tsi.enst.fr/~cappe/h2m/

% Input args.
error(nargchk(5, 6, nargin));
if (nargin < 6)
  QUIET = 0;
end
[N, p, DIAG_COV] = hmm_chk(A, pi0, mu, Sigma);
[T, Nc] = size(X);
if (Nc ~= p)
  error('Observation vectors have an incorrect dimension.');
end

% Compute density values
dens = gauseval(X, mu, Sigma, QUIET);

% Forward pass
if (~QUIET)
  fprintf(1, 'Forward filtering...'); time = cputime;
end
normc = [1 ; zeros(T-1, 1)];
gamma = zeros(T, N);
gamma(1,:) = pi0.*dens(1,:);
for i=2:T
  gamma(i,:) = (gamma(i-1,:) * A) .* dens(i,:);
  normc(i) = sum(gamma(i,:));
  gamma(i,:) = gamma(i,:) / normc(i);
end
% This is not computationnaly efficient but log(prod()) won't work here !
logl = sum(log(normc));
if (~QUIET)
  time = cputime - time;
  fprintf(1, ' (%.2f s)\n', time);
end

% Backward pass
if (~QUIET)
  fprintf(1, 'Backward smoothing...'); time = cputime;
end
% Below, K is the backward smoothing transition matrix
Ahat = zeros(N);
for i=(T-1):-1:1
  K = gamma(i*ones(N,1),:) .* A';
  K = K ./ (sum(K,2)*ones(1,N)); % Requires MATLAB V 5
  TMP = (gamma(i+1,:)' * ones(1,N)) .* K;
  % Beware: TMP is the joint dist. with row index corresponding to T+1
  % and column index to T
  Ahat = Ahat + TMP';
  gamma(i,:) = sum(TMP);
end
% Normalization of the transition matrix
Ahat = Ahat ./ (sum(Ahat.').' *ones(1,N));
if (~QUIET)
  time = cputime - time; fprintf(1, ' (%.2f s)\n',time);
end
